@@ -6,21 +6,23 @@ import os
 from functools import partial
 from multiprocessing import Pool
 from pathlib import Path
-from typing import List, Union, Tuple, Optional, Type
+from typing import List, Union, Tuple, Optional, Type, Dict, Any
+from typing import TYPE_CHECKING
 
 import networkx as nx
 import numpy as np
-from ipyleaflet import Map, Polyline, Marker, Icon, FullScreenControl, ScaleControl
-from ipywidgets import HTML
 from networkx import connected_components
 from tqdm.auto import tqdm
 
 from . import config
 from .file import RDYFile
-from .osm import OSM, OSMRailwaySwitch, OSMRailwaySignal, OSMLevelCrossing
+from .osm import OSM
 from .osm.utils import boxes_to_edges, iou
 from .utils import GPSSeries, TimeSeries
 
+if TYPE_CHECKING:
+    from .widgets import Map
+
 logger = logging.getLogger(__name__)
 
 
@@ -32,7 +34,7 @@ class Campaign:
                  sync_method: str = "timestamp",
                  timedelta_unit: str = 'timedelta64[ns]',
                  strip_timezone: bool = True,
-                 cutoff: bool = True,
+                 trim_ends: bool = True,
                  lat_sw: float = None,
                  lon_sw: float = None,
                  lat_ne: float = None,
@@ -58,7 +60,7 @@ class Campaign:
             Method to use to sync timestamps of individual files
         strip_timezone: bool, default: True
             Strips timezone from timestamps as np.datetime64 does not support timezones
-        cutoff: bool, default: True
+        trim_ends: bool, default: True
             If True, cutoffs the measurements precisely to the timestamp when the measurement was started, respectively
             stopped. By default, Ridy measurement files can contain several seconds of measurements from before/after
             the button press
@@ -85,6 +87,7 @@ class Campaign:
         self.folder = folder
         self.name = name
         self.files: List[RDYFile] = []
+        self.grouped_files: Dict[Any, RDYFile] = {}
 
         # Geographic extent of campaign
         self.lat_sw, self.lon_sw = lat_sw, lon_sw
@@ -124,13 +127,13 @@ class Campaign:
         self.sync_method = sync_method
         self.timedelta_unit = timedelta_unit # Only relevant for timestamp sync method
         self.strip_timezone = strip_timezone
-        self.cutoff = cutoff
+        self.trim_ends = trim_ends
 
         self.results = {}  # Dictionary for Post Processing Results
 
         if folder:
             self.import_folder(self.folder, recursive, exclude,
-                               cutoff=self.cutoff,
+                               trim_ends=self.trim_ends,
                                sync_method=self.sync_method,
                                timedelta_unit=self.timedelta_unit,
                                strip_timezone=self.strip_timezone)
@@ -162,163 +165,13 @@ class Campaign:
 
     @osm.setter
     def osm(self, value):
-        for f in tqdm(self):
+        for f in tqdm(self, desc="OSM Setter (Files in Campaign)"):
             f.osm = value
             if self.map_matching:
                 f.do_map_matching()
 
         self._osm = value
 
-    def add_tracks_to_map(self, m: Map) -> Map:
-        """ Add all GPS tracks from the campaign files to a Map
-
-        Parameters
-        ----------
-        m: Map
-            ipyleaflet Map
-
-        Returns
-        -------
-        Map
-
-        """
-        for file in self.files:
-            m = self.add_track_to_map(m, file=file)
-
-        return m
-
-    def add_track_to_map(self, m: Map, name: str = "", file: RDYFile = None) -> Map:
-        """ Adds a GPS track from a file to the Map
-
-        Parameters
-        ----------
-        m: Map
-            ipyleaflet map
-        name: str
-            Name of the file that should be drawn onto the map
-        file: RDYFile
-            Alternatively, provide RDYFile that should be drawn on the map
-
-        Returns
-        -------
-        Map
-
-        """
-        if name != "":
-            files = [self(name)]
-        elif file is not None:
-            files = [file]
-
-        else:
-            raise ValueError("You must provide either a filename or the file")
-
-        for f in files:
-            gps_series = f.measurements[GPSSeries]
-            coords = gps_series.to_ipyleaflef()
-
-            if coords != [[]]:
-                file_polyline = Polyline(locations=coords, color=f.color, fill=False, weight=4,
-                                         dash_array='10, 10')
-                m.add_layer(file_polyline)
-
-                # Add Start/End markers
-                start_icon = Icon(
-                    icon_url='https://raw.githubusercontent.com/pointhi/leaflet-color-markers/master/img/marker-icon-green.png',
-                    shadow_url='https://cdnjs.cloudflare.com/ajax/libs/leaflet/0.7.7/images/marker-shadow.png',
-                    icon_size=[25, 41],
-                    icon_anchor=[12, 41],
-                    popup_anchor=[1, -34],
-                    shadow_size=[41, 41])
-
-                end_icon = Icon(
-                    icon_url='https://raw.githubusercontent.com/pointhi/leaflet-color-markers/master/img/marker-icon-red.png',
-                    shadow_url='https://cdnjs.cloudflare.com/ajax/libs/leaflet/0.7.7/images/marker-shadow.png',
-                    icon_size=[25, 41],
-                    icon_anchor=[12, 41],
-                    popup_anchor=[1, -34],
-                    shadow_size=[41, 41])
-
-                start_marker = Marker(location=tuple(coords[0]), draggable=False, icon=start_icon)
-                end_marker = Marker(location=tuple(coords[-1]), draggable=False, icon=end_icon)
-
-                start_message = HTML()
-                end_message = HTML()
-                start_message.value = "<p>Start:</p><p>" \
-                                      + str(f.filename or '') + "</p><p>" \
-                                      + str(getattr(f.device, "manufacturer", "")) + "; " \
-                                      + str(getattr(f.device, "model", "")) + "</p>"
-                end_message.value = "<p>End:</p><p>" \
-                                    + str(f.filename or '') + "</p><p>" \
-                                    + str(getattr(f.device, "manufacturer", "")) + "; " \
-                                    + str(getattr(f.device, "model", "")) + "</p>"
-
-                start_marker.popup = start_message
-                end_marker.popup = end_message
-
-                m.add_layer(start_marker)
-                m.add_layer(end_marker)
-
-        return m
-
-    def add_osm_routes_to_map(self, m: Map) -> Map:
-        """ Adds OSM Routes from the downloaded OSM Region
-
-        Parameters
-        ----------
-        m: Map
-            ipyleaflet Map
-
-        Returns
-        -------
-        Map
-
-        """
-        if self.osm:
-            for line in self.osm.railway_lines:
-                for track in line.tracks:
-                    coords = track.to_ipyleaflet()
-                    file_polyline = Polyline(locations=coords, color=line.color, fill=False, weight=4)
-                    m.add_layer(file_polyline)
-        else:
-            logger.warning("No OSM region downloaded!")
-
-        return m
-
-    def add_osm_railway_elements_to_map(self, m: Map) -> Map:
-        """ Draws railway elements using markers on top of a map
-
-        Parameters
-        ----------
-        m: Map
-            Map where railway elements should be drawn onto
-
-        Returns
-        -------
-        m: Map
-            Map containing railway elements
-
-        """
-        if self.osm:
-            for el in self.osm.railway_elements:
-                if type(el) == OSMRailwaySwitch:
-                    icon = Icon(
-                        icon_url='https://raw.githubusercontent.com/pointhi/leaflet-color-markers/master/img/marker-icon-black.png',
-                        shadow_url='https://cdnjs.cloudflare.com/ajax/libs/leaflet/0.7.7/images/marker-shadow.png',
-                        icon_size=[25, 41],
-                        icon_anchor=[12, 41],
-                        popup_anchor=[1, -34],
-                        shadow_size=[41, 41])
-                    marker = Marker(location=(el.lat, el.lon), draggable=False, icon=icon)
-
-                    m.add_layer(marker)
-                elif type(el) == OSMRailwaySignal:
-                    pass
-                elif type(el) == OSMLevelCrossing:
-                    pass
-                else:
-                    pass
-        return m
-
     def clear_files(self):
         """
             Clear all files from the campaign
@@ -327,8 +180,8 @@ class Campaign:
 
     def create_map(self, center: Tuple[float, float] = None,
                    show_gps_tracks=True,
-                   show_railway_elements=False) -> Map:
-        """ Creates a ipyleaflet map showing the GPS tracks of measurement files
+                   show_railway_elements=False) -> 'Map':
+        """ Creates a pyridy.widgets Map (based on ipyleaflet) showing the GPS tracks of measurement files
 
         Parameters
         ----------
@@ -338,31 +191,25 @@ class Campaign:
 
         Returns
         -------
-
+        pyridy.widgets Map
         """
+        from .widgets import Map
+
         if not center:
-            if self.lat_sw and self.lat_ne and self.lon_sw and self.lon_ne:
-                center = (
-                    (self.lat_sw + self.lat_ne) / 2,
-                    (self.lon_sw + self.lon_ne) / 2)
-            else:
-                raise ValueError("Cant determine geographic center of campaign, enter manually using 'center' argument")
+            center = self.determine_geographic_center()
 
-        m = Map(center=center, zoom=12, scroll_wheel_zoom=True, basemap=config.OPEN_STREET_MAP_DE)
-        m.add_control(ScaleControl(position='bottomleft'))
-        m.add_control(FullScreenControl())
+        m = Map(center=center, zoom=12)
 
-        # Add map
-        m.add_layer(config.OPEN_RAILWAY_MAP)
-
-        # Plot GPS point for each measurement and OSM Tracks
-        m = self.add_osm_routes_to_map(m)
+        # Plot OSM Tracks
+        m.osm_routes_layer = m.add_osm_routes(self)
 
+        # Plot measurement GPS Tracks
         if show_gps_tracks:
-            m = self.add_tracks_to_map(m)
+            m.measurement_layers = m.add_measurements(self)
 
+        # Plot railway elements
         if show_railway_elements:
-            m = self.add_osm_railway_elements_to_map(m)
+            m.railway_elements_layer = m.add_osm_railway_elements(self)
 
         return m
 
@@ -395,6 +242,14 @@ class Campaign:
         logging.info("Geographic boundaries of measurement campaign: Lat SW: %s, Lon SW: %s, Lat NE: %s, Lon NE: %s"
                      % (str(self.lat_sw), str(self.lon_sw), str(self.lat_ne), str(self.lon_ne)))
 
+    def determine_geographic_center(self) -> Tuple[float, float]:
+        if self.lat_sw and self.lat_ne and self.lon_sw and self.lon_ne:
+            return (
+                (self.lat_sw + self.lat_ne) / 2,
+                (self.lon_sw + self.lon_ne) / 2)
+        else:
+            raise ValueError("Cannot determine geographic center of campaign, enter manually using 'center' argument")
+
     def do_map_matching(self, rematch=False, **kwargs):
         """ Performs map matching for all files in campaign
 
@@ -404,7 +259,7 @@ class Campaign:
             If True performs map matching again, even when file already contains a map matching
         """
         if self.osm:
-            for f in tqdm(self):
+            for f in tqdm(self, desc="Map Matching: Files in Campaign"):
                 if f.matched_ways and f.matched_nodes:
                     if rematch:
                         logger.info("(%s) File already has a map matching! Rematching..." % f.filename)
@@ -418,8 +273,32 @@ class Campaign:
             raise RuntimeError("Can't do Map Matching, because no OSM data has been downloaded!")
         pass
 
-    def download_osm_data(self):
+    def group_by(self, key):
+        """ Groupy Ridy files by given key, grouped files can be accessed through the grouped_files attributes
+
+        Parameters
+        ----------
+        key: Any
+            Key that should be used to group files
+        """
+        f_sorted = sorted(self.files, key=key)
+        self.grouped_files = {k: list(it) for k, it in itertools.groupby(f_sorted, key)}
+
+    def determine_bounding_boxes(self, simplify=True, plot=False):
+        """
+        Determine bounding boxes of files.
+        Stores bounding boxes in self.bboxs.
+
+        If `simplify=True`:
+        Unify bounding boxes with a large overlap to reduce number of queries and
+        stores these in self.s_bboxs.
+        """
+        # determine bounding boxes of files
         self.bboxs = [f.bbox for f in self.files if f.bbox]
+        if simplify:
+            self.simplify_bounding_boxes()
+
+    def simplify_bounding_boxes(self):
         self.s_bboxs = []  # Filtered bounding boxes
 
         # Unify bounding boxes with a large overlap to reduce number of queries
@@ -460,22 +339,22 @@ class Campaign:
         # ax.set_ylim([self.lat_sw, self.lat_ne])
         # plt.show()
 
+    def download_osm_data(self):
         if config.options["OSM_SINGLE_BOUNDING_BOX"]:
             self.osm = OSM(bbox=self.extent, desired_railway_types=self.railway_types, recurse=self.osm_recurse_type)
         else:
-            if config.options["OSM_BOUNDING_BOX_OPTIMIZATION"] and self.s_bboxs:
+            self.determine_bounding_boxes(simplify=False)
+            if config.options["OSM_BOUNDING_BOX_OPTIMIZATION"]:
+                self.simplify_bounding_boxes()
                 self.osm = OSM(bbox=self.s_bboxs, desired_railway_types=self.railway_types,
                                recurse=self.osm_recurse_type)
             else:
-                if self.bboxs:
-                    self.osm = OSM(bbox=self.bboxs, desired_railway_types=self.railway_types, recurse=self.osm_recurse_type)
-                else:
-                    raise ValueError("Can't retrieve OSM Data because bounding boxes are empty!")
+                self.osm = OSM(bbox=self.bboxs, desired_railway_types=self.railway_types, recurse=self.osm_recurse_type)
 
     def import_files(self, file_paths: Union[list, str] = None,
                      sync_method: str = "timestamp",
                      timedelta_unit: str = 'timedelta64[ns]',
-                     cutoff: bool = True,
+                     trim_ends: bool = True,
                      strip_timezone: bool = True,
                      det_geo_extent: bool = True,
                      use_multiprocessing: bool = False,
@@ -492,8 +371,8 @@ class Campaign:
             Timedelta unit for timestamp sync method
         strip_timezone: bool, default: True
             If True, strips timezone from timestamp arrays
-        cutoff: bool, default: True
-            If True, cutoffs measurement precisely to timestamp when the measurement was started respectively stopped
+        trim_ends: bool, default: True
+            If True, trims measurement precisely to timestamp when the measurement was started respectively stopped
         file_paths: str or list of str
             Individual file paths of the files that should be imported
         sync_method: str
@@ -512,6 +391,9 @@ class Campaign:
         if osm_recurse_type:
             self.osm_recurse_type = osm_recurse_type
 
+        if railway_types:
+            self.railway_types = railway_types
+
         if type(file_paths) == str:
             file_paths = [file_paths]
         elif type(file_paths) == list:
@@ -537,21 +419,20 @@ class Campaign:
                 files = list(tqdm(p.imap(partial(RDYFile,
                                                  sync_method=sync_method,
                                                  strip_timezone=strip_timezone,
-                                                 cutoff=cutoff,
-                                                 series=self._series), file_paths)))
+                                                 trim_ends=trim_ends,
+                                                 series=self._series), file_paths),
+                                  desc="multiprocessing pool"))
                 for f in files:
                     self.files.append(f)
         else:
-            for p in tqdm(file_paths):
+            for p in tqdm(file_paths, desc="file paths"):
                 self.files.append(RDYFile(path=p,
                                           sync_method=sync_method,
                                           timedelta_unit=timedelta_unit,
                                           strip_timezone=strip_timezone,
-                                          cutoff=cutoff,
+                                          trim_ends=trim_ends,
                                           series=self._series))
 
-        self.railway_types = railway_types
-
         if osm_recurse_type:
             self.osm_recurse_type = osm_recurse_type
 
